import Foundation
import PlaygroundSupport


/**
Oberklasse aller Bausteine des Gartens

- author: Albert Wiedemann
- version: 1.0
*/
class GartenElement
{
    /** Die Seitenlänge eines Elementquadrats */
    static let seitenlänge = 20
    /** Das Darstellungsrechteck */
    fileprivate var darstellung: Rechteck
    /** x-Position des Quadrats in Zellenangben */
    fileprivate var x: Int
    /** y-Position des Quadrats in Zellenangben */
    fileprivate var y: Int

    /**
	Legt das Darstellungsrechteck an.
	- parameters:
		- x: x-Position des Quadrats in Zellenangben
		- y: y-Position des Quadrats in Zellenangben
    */
    init(x: Int, y: Int)
    {
        self.x = x
        self.y = y
        darstellung = Rechteck()
        darstellung.PositionSetzen(x: x * GartenElement.seitenlänge, y: y * GartenElement.seitenlänge)
        darstellung.GrößeSetzen(breite: GartenElement.seitenlänge, höhe: GartenElement.seitenlänge)
    }
    
    /**
    Gibt die x-Position des Elements zurück.
	- returns: x-Position
    */
    func XPositionGeben() -> Int
    {
        return x
    }
    
    /**
	Gibt die y-Position des Elements zurück.
	- returns: y-Position
     */
    func YPositionGeben() -> Int
    {
        return y
    }
}

/**
Ein Rasenstück des Gartens

- author: Albert Wiedemann
- version: 1.0
*/
class Rasen: GartenElement
{

    /**
	Legt das Darstellungsrechteck in der Rasenfarbe an.
	- parameters:
		- x: x-Position des Quadrats in Zellenangben
		- y: y-Position des Quadrats in Zellenangben
    */
    override init(x: Int, y: Int)
    {
        super.init(x: x, y: y)
        darstellung.FarbeSetzen(farbe: "grün")
    }
}

/**
Ein Hindernis des Gartens

- author: Albert Wiedemann
- version: 1.0
*/
class Hindernis: GartenElement
{

    /**
	Legt das Darstellungsrechteck in der Hindernisfarbe an.
	- parameters:
		- x: x-Position des Quadrats in Zellenangben
		- y: y-Position des Quadrats in Zellenangben
    */
    override init(x: Int, y: Int)
    {
        super.init(x: x, y: y)
        darstellung.FarbeSetzen(farbe: "braun")
    }
}

/**
Verwaltet den Garten

- author: Albert Wiedemann
- version: 1.0
*/
class Garten
{
    /** Feld der Gartenelemente */
    private var elemente: [[GartenElement]]

    /**
    Legt den Garten an.
	- parameters:
		- dateiName: Name der Datei mit der Beschreibung des Gartens
		- graph: der Graph, der den Garten repräsentiert.
     */
    init(dateiName: String, graph: GraphMatrix)
    {
        elemente = [[GartenElement]]()
        if Lesen(dateiName: dateiName)
        {
            GraphErzeugen(graph: graph)
        }
    }
    
    /**
	Liest die Beschreibungs des Gartens aus einer Datei ein
	- parameters:
		- dateiName: Name der Datei
	- returns: wahr, wenn die Datei gelesen werden konnte
     */
    private func Lesen (dateiName: String) -> Bool
    {
        do
        {
            guard let datei = Bundle.main.url(forResource: dateiName, withExtension: "")
            else
            {
            	print ("Die Datei \(dateiName) exisitert nicht")
            	return false
            }
            let zeilen = try String(contentsOf: datei, encoding: .utf8).split(separator: "\n")
            var y = 0
            for zeile in zeilen
            {
                elemente.append([GartenElement]())
                
                var x = 0
                for (_, ch) in zeile.lowercased().enumerated()
                {
                    var e: GartenElement
                    if ch == "h"
                    {
                        e = Hindernis(x: x, y: y)
                    }
                    else if ch == "r"
                    {
                        e = Rasen(x: x, y: y)
                    }
                    else
                    {
                        e = GartenElement(x: x, y: y)
                        e.darstellung.FarbeSetzen(farbe: "rot")
                    }
                    elemente[y].append(e)
                    x += 1
                }
                y += 1
            }
            return true
        }
        catch
        {
            print("Fehler beim Lesen der Datei: \(error)")
            return false
        }
    }
    
    /**
	Erzeugt die Graphendarstellung
	- parameters:
		- graph: das Objekt zur Verwaltung des Graphen
     */
    func GraphErzeugen(graph: GraphMatrix)
    {
        for y in 0 ..< elemente.count
        {
            for x in 0 ..< elemente[y].count
            {
                if elemente[y][x] is Rasen
                {
                    graph.KnotenEinfügen(x: x, y: y)
                    if (y > 0) && (elemente[y - 1][x] is Rasen)
                    {
                        graph.KanteEinfügen(x1: x, y1: y - 1, x2: x, y2: y)
                    }
                    if (x > 0) && (elemente[y][x - 1] is Rasen)
                    {
                        graph.KanteEinfügen(x1: x - 1, y1: y, x2: x, y2: y)
                    }
                }
            }
        }
    }
}

/**
Verwaltet ein Knotensymbol in einem Graph

- author: Albert Wiedemann
- version: 1.0
*/
class KnotenSymbol
{
    /** Außenkreis des Darstellungssymbols */
    private let außen: Kreis
    /** Innenkreis des Darstellungssymbols */
    private let innen: Kreis
    /** Text des Darstellungssymbols */
    private let text: Text
    /** Der Bezeichner */
    private var bezeichner: String
    /** Radius */
    private var r: Int
    /** x-Koordinate */
    private var x: Int
    /** y-Koordinate */
    private var y: Int
    /** Farbe */
    private var farbe: String
    
    /**
	Legt das Symbol an und besetzt die Attribute.
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
		- radius: Radius
		- farbe: Farbe
		- bezeichner: Bezeichner
    */
    init(x: Int, y: Int, radius: Int, farbe: String, bezeichner: String)
    {
        self.x = x
        self.y = y
        self.r = radius
        self.farbe = farbe
        self.bezeichner = bezeichner
        außen = Kreis()
        außen.FarbeSetzen(farbe: "schwarz")
        innen = Kreis()
        text = Text()
        text.TextGrößeSetzen(größe: 18)
        DarstellungAktualisieren()
    }
    
    /**
    Aktualisiert das Symbol
    */
    private func DarstellungAktualisieren()
    {
        außen.PositionSetzen(x: x, y: y)
        außen.RadiusSetzen(radius: r)
        innen.PositionSetzen(x: x, y: y)
        innen.RadiusSetzen(radius: r - 2)
        innen.FarbeSetzen(farbe: farbe)
        text.PositionSetzen(x: x - bezeichner.count / 2 * 10 - bezeichner.count % 2 * 5, y: y + 6)
        text.TextSetzen(text: bezeichner)
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        DarstellungAktualisieren()
    }
    
    /**
	Setzt die Position der Darstellung
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
    */
    func PositionSetzen(x: Int, y: Int)
    {
        self.x = x
        self.y = y
        DarstellungAktualisieren()
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- bezeichner: der (neue) Bezeichner
    */
    func BezeichnerSetzen(bezeichner: String)
    {
        self.bezeichner = bezeichner
        DarstellungAktualisieren()
    }
    
    /**
	Meldet die x-Koordinate des Symbols zurück.
	- returns: x-Koordinate
    */
    func XGeben() -> Int
    {
        return x
    }
    
    /**
	Meldet die y-Koordinate des Symbols zurück.
	- returns: y-Koordinate
     */
    func YGeben() -> Int
    {
        return y
    }
    
    /**
	Meldet den Bezeichner des Symbols zurück.
	- returns: Bezeichner
     */
    func BezeichnerGeben() -> String
    {
        return bezeichner
    }
    
    /**
    Entfernt das Knotensymbol aus der Anzeige
    */
    func Entfernen()
    {
        außen.Entfernen()
        innen.Entfernen()
        text.Entfernen()
    }
}

/**
Verwaltet ein Kantensymbol in einem Graph

- author: Albert Wiedemann
- version: 1.0
*/
class KantenSymbol
{
    /** Außenrechteck des Darstellungssymbols */
    private let außen: Rechteck
    /** Innenrechteck des Darstellungssymbols */
    private let innen: Rechteck
    /** Richtungspfeil des Darstellungssymbols */
    private let pfeil: Dreieck
    /** Gewichtsangabe des Darstellungssymbols */
    private let text: Text
    /** Breite des Darstellungssymbols */
    private var breite: Int
    /** Gerichtete Kante */
    private var gerichtet: Bool
    /** Gewicht der Kante */
    private var gewicht: String
    /** Farbe */
    private var farbe: String
    /** Startknoten */
    private var von: KnotenSymbol
    /** Endknoten */
    private var nach: KnotenSymbol

    /**
	Legt das Symbol an und besetzt die Attribute.
	- parameters:
		- start: Startknoten
		- ende: Endknoten
		- gerichtet: wenn wahr, ist der Weg gerichtet
		- gewicht: Kantengewicht
		- breite: Breite der Linie
		- farbe: Farbe der Linie
    */
    init(start: KnotenSymbol, ende: KnotenSymbol, gerichtet: Bool, gewicht: String, breite: Int, farbe: String)
    {
        von = start
        nach = ende
        self.farbe = farbe
        self.breite = breite
        self.gerichtet = gerichtet
        self.gewicht = gewicht
        außen = Rechteck()
        innen = Rechteck()
        pfeil = Dreieck()
        pfeil.FarbeSetzen(farbe: "schwarz")
        text = Text()
        text.SichtbarkeitSetzen(sichtbar: !("" == gewicht))
        text.TextSetzen(text: gewicht)
        DarstellungAktualisieren()
    }
    
    /**
    Aktualisiert das Symbol
    */
    func DarstellungAktualisieren()
    {
        let x1 =  Double(von.XGeben())
        let y1 = Double(von.YGeben())
        let x2 = Double(nach.XGeben())
        let y2 = Double(nach.YGeben())
        let länge = Int(sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)))
        let xm = Int((x1 + x2) / 2.0)
        let ym = Int((y1 + y2) / 2.0)
        let x = xm - länge / 2
        let y = ym - breite / 2
        außen.PositionSetzen(x: x, y: y)
        außen.GrößeSetzen(breite: länge, höhe: breite)
        if breite < 4
        {
            let größe = breite * 4
            innen.SichtbarkeitSetzen(sichtbar: false)
            außen.FarbeSetzen(farbe: farbe)
            pfeil.SichtbarkeitSetzen(sichtbar: gerichtet)
            pfeil.PositionSetzen(x: xm, y: ym - größe / 2)
            pfeil.GrößeSetzen(breite: größe, höhe: größe)
            pfeil.WinkelSetzen(winkel: 270)
        }
        else
        {
            let delta = breite <= 8 ? 2 : 4
            innen.SichtbarkeitSetzen(sichtbar: true)
            innen.PositionSetzen(x: x, y: y + delta / 2)
            innen.GrößeSetzen(breite: länge, höhe: breite - delta)
            innen.FarbeSetzen(farbe: farbe)
            außen.FarbeSetzen(farbe: "schwarz")
            pfeil.SichtbarkeitSetzen(sichtbar: gerichtet)
            pfeil.PositionSetzen(x: xm, y: ym - breite / 2)
            pfeil.GrößeSetzen(breite: breite - delta, höhe: breite)
            pfeil.WinkelSetzen(winkel: 270)
        }
        
        var winkel = 0.0
        if x1 == x2
        {
            if y2 > y1
            {
                winkel = -90
            }
            else
            {
                winkel = 90
            }
        }
        else if x1 < x2
        {
            winkel = -atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }
        else
        {
            winkel = 180 - atan((y2 - y1) / (x2 - x1)) / Double.pi * 180.0
        }
        pfeil.WinkelSetzen(winkel: Int(winkel) + 270)
        innen.WinkelSetzen(winkel: Int(winkel))
        außen.WinkelSetzen(winkel: Int(winkel))
        pfeil.GanzNachHintenBringen()
        innen.GanzNachHintenBringen()
        außen.GanzNachHintenBringen()
        if (-90 <= winkel) && (winkel <= 0)
        {
            let w = (winkel + 90) / 180.0 * Double.pi
            let delta = Double(breite)
            text.PositionSetzen(x: xm + Int(delta * cos(w)), y: ym - Int(delta * sin(w)))
        }
        else if (0 < winkel) && (winkel <= 90)
        {
            let w = (winkel + 90) / 180.0 * Double.pi
            let delta = Double(breite)
            text.PositionSetzen(x: xm - Int(delta * cos(w)), y: ym + Int(delta * sin(w)) + 7)
        }
        else if (90 < winkel) && (winkel <= 180)
        {
            let w = (winkel + 90) / 180.0 * Double.pi
            let delta = Double(breite)
            text.PositionSetzen(x: xm - Int(delta * cos(w)), y: ym + Int(delta * sin(w)))
        }
        else
        {
            let w = (winkel + 90) / 180.0 * Double.pi
            let delta = Double(breite)
            text.PositionSetzen(x: xm + Int(delta * cos(w)), y: ym - Int(delta * sin(w)) + 10)
        }
    }
    
    /**
    Entfernt das Kantensymbol aus der Anzeige
    */
    func Entfernen()
    {
        außen.Entfernen()
        innen.Entfernen()
        pfeil.Entfernen()
        text.Entfernen()
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        farbe = f
        DarstellungAktualisieren()
    }
    
    /**
	Meldet den Startknoten
	- returns: Startknoten
    */
    func StartsymbolGeben() ->KnotenSymbol
    {
        return von
    }
    
    /**
	Meldet den Zielknoten
	- returns: Zielknoten
    */
    func ZielsymbolGeben() -> KnotenSymbol
    {
        return nach
    }
}

/**
Beschreibt einen Knoten

- author: Albert Wiedemann
- version: 1.0
*/
class Knoten
{
    /** x-Koordinate des Knotens */
    private let x: Int
    /** y-Koordinate des Knotens */
    private let y: Int
    /** Das Darstellungssymbol des Knotens */
    private var symbol: KnotenSymbol
    
    /**
	Besetzt die Attribute und legt das Knotensymbol an.
	- parameters:
		- x: x-Koordinate
		- y: y-Koordinate
    */
    init(x: Int, y: Int)
    {
        self.x = x
        self.y = y
        symbol = KnotenSymbol(x: x * GartenElement.seitenlänge + GartenElement.seitenlänge / 2, y: Int (anzeigeHöhe) / 2 + y * GartenElement.seitenlänge + GartenElement.seitenlänge / 2, radius: (GartenElement.seitenlänge - 4) / 2, farbe: "weiß", bezeichner: "")
    }
    
    /**
	Meldet die x-Koordinate des Knotens zurück.
	- returns: x-Koordinate
    */
    func XGeben() -> Int
    {
        return x
    }
    
    /**
	Meldet die y-Koordinate des Knotens zurück.
	- returns: y-Koordinate
    */
    func YGeben() -> Int
    {
        return y
    }
    
    /**
	Meldet das Symbol des Knotens zurück.
	- returns: Knotensymbol
    */
    func SymbolGeben() -> KnotenSymbol
    {
        return symbol
    }
    
    /**
	Setzt die Farbe der Darstellung
	- parameters:
		- f: die (neue) Farbe
    */
    func FarbeSetzen(f: String)
    {
        symbol.FarbeSetzen(f: f)
    }
}

/**
Verwaltet einen ungerichteten, gewichteten Graphen mittels Adjazenzmatrix

- author: Albert Wiedemann
- version: 1.0
*/
class GraphMatrix
{
    /** Feld der Knoten des Graphen */
    private var knoten: [Knoten]
    /** 2-dim Feld der Adjazenzmatrix */
    private var matrix: [[Int]]
    /** Feld der Kantensymbole des Graphen */
    private var kanten: [KantenSymbol]
    
    /**
    Baut die Datenstruktur auf
    */
    init()
    {
        knoten = [Knoten]()
        matrix = [[Int]]()
        kanten = [KantenSymbol]()
    }
    
    /**
	Einfügen eines neuen Knoten in den Graphen
	- parameters:
		- x: x-Koordinate für die Anzeige des Knotens
		- y: y-Koordinate für die Anzeige des Knotens
    */
    func KnotenEinfügen(x: Int, y: Int)
    {
        knoten.append(Knoten(x: x, y: y))
        for index in 0 ..< matrix.count
        {
            matrix[index].append(-1)
        }
        matrix.append([Int]())
        for _ in 0 ..< matrix.count
        {
            matrix[matrix.count - 1].append(-1)
        }
    }

    /**
	Gibt die interne Nummer des Knotens
	Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird -1 zurückgegeben
	- parameters:
		- x: x-Koordinate für die Anzeige des Knotens
		- y: y-Koordinate für die Anzeige des Knotens
	- returns: Indexnummer des Knotens im Knotenfeld; 0 <= res < knoten.count bzw. -1
    */
    func KnotenNummerGeben(x: Int, y: Int) -> Int
    {
        for index in 0 ..< knoten.count
        {
            if (knoten[index].XGeben() == x) && (knoten[index].YGeben() == y)
            {
                return index
            }
        }
        return -1
    }

    /**
	Gibt die interne Nummer des Knotens
	Wenn ein Knoten mit diesem Bezeichner nicht bekannt ist wird -1 zurückgegeben
	- parameters:
		- k: der Knoten
	- returns: Indexnummer des Knotens im Knotenfeld; 0 <= res < knoten.count bzw. -1
    */
    func KnotenNummerGeben(k: Knoten) -> Int
    {
        for index in 0 ..< knoten.count
        {
            if knoten[index] === k
            {
                return index
            }
        }
        return -1
    }

    /**
	Einfügen einer Kante in den Graphen
	Eine Kante ist durch einen Anfangsknoten und einen Endknoten festgelegt und hat eine Gewichtung
	- parameters:
		- x1: x-Koordinate des Anfangsknotens
		- y1: y-Koordinate des Anfangsknotens
		- x2: x-Koordinate des Endknotens
		- y2: y-Koordinate des Endknotens
    */
    func KanteEinfügen(x1: Int, y1: Int, x2: Int, y2: Int)
    {
        let vonNummer = KnotenNummerGeben(x: x1, y: y1)
        let nachNummer = KnotenNummerGeben(x: x2, y: y2)
        if (vonNummer != -1) && (nachNummer != -1) && (vonNummer != nachNummer)
        {
            matrix[vonNummer][nachNummer] = 1
            matrix[nachNummer][vonNummer] = 1
            kanten.append(KantenSymbol(start: knoten[vonNummer].SymbolGeben(), ende: knoten[nachNummer].SymbolGeben(), gerichtet: false, gewicht: "", breite: 3, farbe: "blau"))
        }
    }

    /**
	Gibt die Anzahl der Knoten des Graphen zurück
	- returns:  Anzahl der Knoten
    */
    func KnotenAnzahlgeben() -> Int
    {
        return knoten.count
    }
    
    /**
	Führt die Breitensuche zum Testen des Zusammenhangs durch
	Falls am Ende alle Knoten besucht wurden, ist der Graph zusammenhängend.
	- returns: WAHR, wenn der Graph zusammenhängend ist.
    */
    func ZusammenhangTesten() -> Bool
    {
    	//Nach Aufgabenstellung ergänzen
        return false
    }
}


/**
Rahmenklasse für die Wegesuche aus dem Labyrith.

- author: Albert Wiedemann
- version: 1.0
*/
class Rahmen
{
    /** Der Garten */
    private let garten: Garten
    /** Der zugehörige Graph */
    private let g: GraphMatrix

    /**
	Legt die benötigten Objekte an.
	- parameters:
		- dateiName: Name der Datei mit der Beschreibung
    */
    init(dateiName: String)
    {
        g = GraphMatrix()
        garten = Garten(dateiName: dateiName, graph: g)
    }
    
    /**
	Prüft, ob der Graph zusammenhängend ist und gibt eine entsprechende Meldung aus
    */
    func ZusammenhangPrüfen()
    {
        //Nach Aufgabenstellung ergänzen
    }
}

Rahmen(dateiName: "Rasen1.txt").ZusammenhangPrüfen()
//Rahmen(dateiName: "Rasen2.txt").ZusammenhangPrüfen()

